package sdlip.examples;
//
// Encyclopedia Britannica wrapper
//
// This wrapper provides a SDLIP interface to Encyclopedia Britannica
// (www.eb.com)
//
// Author: Arturo Crespo
// Creation Date: May 20, 2000
//
//
// Notes: DID means Document Identifier
//

//package sdlip.examples;

import sdlip.SDLIPException;

import sdlip.SDLIP;
import sdlip.helpers.RangeEnumerator;
import sdlip.helpers.NestedAttribute;
import sdlip.helpers.DBG;
import sdlip.xml.XMLEntities;
import sdlip.xml.dom.*;
import org.w3c.dom.*;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This is an auxiliary class internally used by the
 * BritanicaLSP application.
 */

public class Britannica {

  // exported schema of encyclopedia britannica
  public static final String Namespace = "http://www.eb.com/schema#";
  public static final String TITLE = Namespace + "title";
  public static final String ARTICLE = Namespace + "article";
  public static final String ABSTRACT = Namespace + "abstract";

  // debuging manager
  public DBG dbg = new DBG(1);


  /**
   * Constructor
   */
  public Britannica(String query_) {
    query = (query_!=null? query_ : "");
  }


  public void getDocs(Element eSearchResult, RangeEnumerator range, Element eDocProps) throws SDLIPException {

    // get the list of attributes to retrieve
    Vector attrs = NestedAttribute.collectAttributes(eDocProps);

    // we iterate over the range until we are done (or we are interrupted)
    while(!interrupted && range.nextRange()) {
      int start = range.get();
      int end;
      if (range.limit() == RangeEnumerator.INFINITY) {
	end = 100; // we return up to 100 articles when requested all articles
      } else {
	end = (range.limit()>=range.get()? range.limit()-range.get()+1 : 1);
      }      
	
      String rawSourceOutput = doSourceLookup( start, end );
      parseResult(eSearchResult, attrs, range.get(), rawSourceOutput);
    }
  }

  public synchronized int getNumDocs() {
    return numDocs;
  }

  public void interrupt() {
    interrupted = true;
  }

  //
  // PRIVATE METHODS START HERE
  //
  // Enter at your own risk
  //

  /*---------------------------
  *doSourceLookup
  *--------------*/

  /** Does the dirty work of invoking the HTTP command
   *  for www.eb.com 
   *  Results are returned from "startFrom" and up to "numResults"
   *  results. 
   */


  private String doSourceLookup(int startFrom, int numResults) {

    // sanity check of arguments
    dbg.println(DBG.SHORT, "doSourceLookup(" + startFrom + ", " +
		numResults + ")");
    if(startFrom < 0)
      startFrom = 0;
    if(numResults < 0)
      numResults = 0;
    
    // Prepare HTTP command
    String com = "http://search.eb.com/bol/search?Dbase=Articles" +
      "&query=" + URLEncoder.encode(query) +
      "&skip=" + (startFrom-1) +
      "&hits=" + numResults;

    dbg.println(DBG.SHORT, "doSourceLookup: com = " + com );

    StringBuffer result = new StringBuffer(4096);

    try {
      // Establish HTTP connecton and get results
      URL url = new URL(com);
      BufferedReader in = new BufferedReader( new InputStreamReader( url.openStream() ));

      char[] buf = new char[1024];
      int len = 0;
    
      while ((len = in.read(buf, 0, 1024)) > 0) {
	result.append(buf, 0, len);
      }

    } catch (Exception e) {
      System.err.println(e);
    }

    dbg.println(DBG.VERBOSE, "doSourceLookup: result = " + result );

    return result.toString();
  }

  /*--------------------------------
  | parseResult
  -------------*/

  /**Given the string with a result returned by the Lookup command,
   *adds to the "Document" the info. 
   */

  private void parseResult( Element eSearchResult, Vector attrs, 
			    int initialDID, String lookupResult) throws SDLIPException { 

    // Check for no records:
    if (lookupResult.indexOf("Sorry, your search yielded no results") > -1) {
      numDocs = 0;
      return; 
    }
    
    // Records found:
    // get the document where we'll put the results
    Document doc = XMLObject.getDocument();

    int start;
    int end;

    // Find out how many records are there:
    start = lookupResult.indexOf("</b> are relevant to your search.");
    if (start == -1) errorParsing();
    start = lookupResult.indexOf("<b>",start-20);
    if (start == -1) errorParsing();
    end = lookupResult.indexOf(" articles",start);
    numDocs = Integer.parseInt(lookupResult.substring(start+3,end));
    dbg.println(DBG.SHORT, "numDocs = " + numDocs);

    // look for start of first name and iterate over all results
    start = lookupResult.indexOf("<wbr><a href");

    while (start != -1) {

      //create new element to put the current result
      Element eDoc = doc.createElement(SDLIP.doc);
      Element ePropList = doc.createElement(SDLIP.propList);
      eDoc.appendChild(ePropList);

      // assign DID
      Element eDID = DOMUtil.createTextNode(doc, SDLIP.DID, String.valueOf(initialDID++));
      eDoc.appendChild(eDID);
	  
      // find article reference
      start +=13;
      end = lookupResult.indexOf(">",start);
      if (end == -1) errorParsing();
      String articleReference = "http://search.eb.com" + 
	lookupResult.substring(start,end);
      ePropList.appendChild( DOMUtil.createTextNode(doc, ARTICLE, articleReference) );
      dbg.println(DBG.VERBOSE, "ARTICLE: " + articleReference);
	
      // find title
      start = end + 1;
      end = lookupResult.indexOf("</a>",start);
      if (end == -1) errorParsing();
      String title = lookupResult.substring(start,end);
      ePropList.appendChild( DOMUtil.createTextNode(doc, TITLE, xmlEntities.normalizeData( title )) );
      dbg.println(DBG.VERBOSE, "TITLE: " + title);

      // find abstract
      start = lookupResult.indexOf("<p><!--html--><p>", end);
      if (start == -1) errorParsing();
      end = lookupResult.indexOf("</a></b></i> <br>",start);
      if (end == -1) errorParsing();
      String summary = lookupResult.substring(start+17,end);
      ePropList.appendChild( DOMUtil.createTextNode(doc, ABSTRACT, xmlEntities.normalizeData( summary )) );
      dbg.println(DBG.VERBOSE, "ABSTRACT: " + summary);

      // match against available data
      // and put element in the result
      eSearchResult.appendChild( NestedAttribute.matchDoc(attrs, eDoc) );

      // find next name (if any)
      start = lookupResult.indexOf("<wbr><a href",start);
    }
    return;
  }


  private void errorParsing() throws SDLIPException {
    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "LSP: could not screen-scrape Encyclopedia Britannica");
  }


  public static void main(String[] args) throws Exception {

    if(args.length == 0) {
      System.err.println("Run BritannicaLSP! This class is used for debugging only.\n" +
			 "Usage: Britannica <query> <range>\n" +
			 "Example: Britannica Vivaldi 1-10");
      System.exit(1);
    }

    Element eSearchResult = XMLObject.getDocument().createElement(SDLIP.SearchResult);
    Britannica b = new Britannica(args[0]);

    b.getDocs(eSearchResult, new RangeEnumerator(args[1]),  null);

    System.out.println( DOMWriter.prettyPrint(eSearchResult) );

  }

  private String query = "";
  private boolean interrupted = false;
  private int numDocs = sdlip.Search.NOT_YET_KNOWN;
  private XMLEntities xmlEntities = new XMLEntities();

}
