package sdlip.examples;

import sdlip.SDLIPException;
/* holder for integer values */
import org.omg.CORBA.IntHolder;
/* transport module libraries */
import sdlip.helpers.*;
/* following packages are imported only if DOM-aware XMLObjects are used */
import sdlip.xml.dom.*;
import org.w3c.dom.*;
/* SDLIP XML element names */
import sdlip.SDLIP;

/**
 * This is an example application included in the SDLIP distribution.
 * <p>Usage: <tt>java sdlip.examples.Client &lt;URL_or_IOR> &lt;method> ...method parameters...</tt>
 * <p>Allows to invoke all methods defined in the Search and ResultAccess interfaces on an LSP using a command line 
 * tool. The <tt>&lt;method></tt> parameter corresponds to the name of the method to be invoked. Following methods are 
 * supported:
 * <p><pre>
 *   search              &lt;clientSID> &lt;queryStr> &lt;numDocs> &lt;stateTimeoutReq>
 *   getSessionInfo      &lt;serverSID>
 *   getDocs             &lt;serverSID> &lt;reqID> &lt;docsToGet>
 *   extendStateTimeout  &lt;serverSID> &lt;additionalTime>
 *   cancelRequest       &lt;serverSID> &lt;reqID>
 * </pre>
 */

public class Client {

  static void bailOut() {
    System.err.println("Usage: java sdlip.examples.Client <server URI> <method> ...method params..\n" +
		       "Following methods are supported:\n" +
		       "  search             <clientSID> <queryStr> <numDocs> <stateTimeoutReq>\n" +
		       "  getSessionInfo     <serverSID>\n" +
		       "  getDocs            <serverSID> <reqID> <docsToGet>\n" +
		       "  extendStateTimeout <serverSID> <additionalTime>\n" +
		       "  cancelRequest      <serverSID> <reqID>"
		       );
    System.exit(1);
  }

  static String expectedTotalString(int total) {
    
    if(total == sdlip.Search.UNKNOWABLE)
      return "unknowable";
    else if(total == sdlip.Search.NOT_YET_KNOWN)
      return "not yet known";
    else
      return String.valueOf(total);
  }

  public static void main(String args[]) {

    try {
      /* extract command line parameters */
      String searchServerURI = args[0];
      String method = args[1]; // search, getSessionInfo, getDocs, extendStateTimeout, cancelRequest

      /* an SDLIP/CORBA or SDLIP/DASL transport module is created
       * depending on whether a URL or an IOR is supplied */
      ClientTransportModule tm = ClientTransportModule.create(searchServerURI);
      tm.setDBG(new DBG(DBG.VERBOSE)); // SILENT));

      if("search".equals(method)) {

	String clientSID = args[2];
	String queryStr = args[3];
	String numDocs = args[4];
	String timeoutReq = args[5];

	/* the result will be held in a DOM-enabled sdlip.xml.dom.XMLObject */
	XMLObject result = new XMLObject();
	IntHolder hTotal = new IntHolder();
	IntHolder hTimeout = new IntHolder();
	IntHolder hServerSID = new IntHolder();
	XMLObject delegate = new XMLObject();

	// if a string is supplied, wrap it into DAV:contains
	queryStr = queryStr.trim();
	if(!queryStr.startsWith("<"))
	  queryStr = "<basicsearch xmlns='DAV:'>" +
	    "  <where><contains>" + queryStr + "</contains></where>" +
	    "</basicsearch>";

	tm.search(Integer.parseInt(clientSID),  /* some client SID */
		  null,            /* search the default subcollection */
		  new XMLObject(queryStr),
		  Integer.parseInt(numDocs),
		  null,            /* retrieve all properties */
		  Integer.parseInt(timeoutReq), /* maintain state */
		  null,            /* no query options */
		  hTotal,          /* expected total number of results */
		  hTimeout,        /* place to return stateTimeout */
		  hServerSID,      /* place to return serverSID - don't care */
		  delegate,        /* place to return serverDelegate - don't care */
		  result           /* place to return search result */
		  );


	System.out.println("RESULT:\n" + result.getPrettyString());
	if(delegate.getElement() != null)
	  System.out.println("DELEGATE:\n" + delegate.getPrettyString());
	System.out.println("EXPECTED TOTAL: " + expectedTotalString(hTotal.value));
	System.out.println("TIMEOUT: " + hTimeout.value);
	System.out.println("SERVER SID: " + hServerSID.value);

      } else if("getSessionInfo".equals(method)) {

	String serverSID = args[2];
	IntHolder hExpectedTotal = new IntHolder();
	IntHolder hStateTimeout = new IntHolder();

	tm.getSessionInfo(Integer.parseInt(serverSID), hExpectedTotal, hStateTimeout);

	System.out.println("EXPECTED TOTAL: " + expectedTotalString(hExpectedTotal.value));
	System.out.println("TIMEOUT: " + hStateTimeout.value);

      } else if("extendStateTimeout".equals(method)) {

	String serverSID = args[2];
	String addTime = args[3];
	IntHolder hTimeAllotted = new IntHolder();

	tm.extendStateTimeout(Integer.parseInt(serverSID),
			      Integer.parseInt(addTime),
			      hTimeAllotted);

	System.out.println("TIME ALLOTTED: " + hTimeAllotted.value);

      } else if("getDocs".equals(method)) {

	String serverSID = args[2];
	String reqID = args[3];
	String docsToGet = args[4];

	XMLObject result = new XMLObject();

	tm.getDocs(Integer.parseInt(serverSID),
		   Integer.parseInt(reqID),
		   null,
		   docsToGet,
		   result);
	System.out.println("RESULT:\n" + result.getPrettyString());

      } else if("cancelRequest".equals(method)) {

	String serverSID = args[2];
	String reqID = args[3];

	tm.cancelRequest(Integer.parseInt(serverSID),
			 Integer.parseInt(reqID));
	
      } else
         bailOut();

    } catch (ArrayIndexOutOfBoundsException arr) {
      bailOut();
    } catch (SDLIPException sdlip) {
      System.err.println("SDLIP exception (" + sdlip.getCode() + "): " + sdlip.getReason());
    }
  }
}
