package sdlip.examples;

import sdlip.SDLIPException;

import sdlip.SDLIP;
import sdlip.helpers.RangeEnumerator;
import sdlip.helpers.NestedAttribute;
import sdlip.xml.dom.*;
import org.w3c.dom.*;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This is an auxiliary class internally used by the
 * SimpleBoardLSP application.
 */

public class SwitchBoard {

  public static final String Namespace = "http://www.switchboard.com/schema#";

  //  public static final String firstName = Namespace + "firstName";
  //  public static final String lastName = Namespace + "lastName";
  public static final String NAME = Namespace + "name";
  public static final String ADDRESS = Namespace + "address";
  public static final String PHONE = Namespace + "phone";
  //  public static final String query = Namespace + "query";

  /**
   * Query format: <query><firstName>Andreas</firstName><lastName>Paepcke</lastName></query>
   */

/**This class invokes the Swithboard web directory service
 *(http:\\www.switchboard.com). 
 *The only public method is lookupName(). Give it a string, and
 *you get a vecotr of ResultDocument instances. Each of these
 *instances describes one record of the result. 
 *
 *Properties of result documents from this search engine are:
 *    name, address, and phone.
 *
 *The input string to lookup name must have the format: 
 *,firstname,lastname,city,state
 *Any prefix of the format is also valid.
 *
 *Background for understanding the implementation:
 * 
 *Request:
 * URL: http://www.switchboard.com/bin/cgiqa.dll
 * Parameters:
 *   - "SR="  <purpose unknown>
 *   - "MEM=1"  <purpose unknown>
 *   - "F=firstname"
 *   - "L=lastname"
 *   - "T=city"
 *   - "S=state"
 *   - "Submit=Search" <purpose unknown>
 *
 *All parameters must be present.  For optional fields (F,T,S), when
 *the value is not known, an empty string must be passed (e.g. "F=").
 *Values must be escaped (white spaces replaced by "+")
 *
 * Example: (looking for John F Kennedy in California)
 *   http://www.switchboard.com/bin/cgiqa.dll?SR=&MEM=1&F=John+F&L=Kennedy&T=&S=CA&Submit=Search
 *
 *Response:
 *There three main responses:
 *
 *1. If name was unique and a record was found:
 *  Identified by string:
 *    "<!-- Area Identification line -->"
 *2. If multiple records with that name were found:
 *  Identified by string:
 *    "Displaying: 1-"
 *  Parsing 
 *3. If no record was found:
 *  Identified by string:
 *    "We are unable to locate any listings that match your search criteria"
 */


  // strings used to determine the type of the document
  static final String oneDocumentStr  = 
    "<!-- Area Identification line -->";
  static final String manyDocumentStr = 
    "Displaying: ";
  static final String noDocumentsStr = 
    "We are unable to locate any listings that match your search criteria";

  String firstName, lastName, cityName, stateName;

  boolean interrupted = false;
  Vector records = new Vector(); // element 0 corresponds to DID 1
  String handle = ""; // last handle from switchboard.com
  
  /**
   * Initialize search
   */
  public SwitchBoard(String firstName, String lastName, String cityName, String stateName) {
    
    this.firstName = firstName;
    this.lastName = lastName;
    this.cityName = cityName;
    this.stateName = stateName;

  }

  public SwitchBoard(String firstName, String lastName) {

    this(firstName, lastName, null, null);
  }

  /**
   * Format: [first-name] last-name
   */
  public SwitchBoard(String query) {

    firstName = "";

    StringTokenizer st = new StringTokenizer(query, " \t");
    while(st.hasMoreTokens()) {
      String s = st.nextToken();
      if(!st.hasMoreTokens())
	lastName = s;
      else {
	if(!"".equals(firstName))
	  firstName += " ";
	firstName += s;
      }
    }
    if(lastName == null)
      throw new RuntimeException("Empty query");
  }

  public void getDocs(Element eSearchResult, RangeEnumerator range, Element eDocProps) throws SDLIPException {

    // get the list of attributes to retrieve
    Vector attrs = NestedAttribute.collectAttributes(eDocProps);

    while(!interrupted && range.next()) {

      while(!interrupted && range.get() > records.size() && handle != null) {
	// fetches more documents
	fetchMore();
      }

      if(range.get() <= records.size()) {
	Element eDoc = (Element)records.elementAt(range.get() - 1);

	// match against available data
	eSearchResult.appendChild( NestedAttribute.matchDoc(attrs, eDoc) );

      } else // nothing left
	break;
    }
  }

  public synchronized int getNumDocs() {
    if(handle != null)
      return sdlip.Search.UNKNOWABLE;
    else
      return records.size();
  }

  public void interrupt() {
    interrupted = true;
  }
  
  private synchronized void fetchMore() throws SDLIPException {
    
    if(handle != null) {
      handle = lookup10(records.size() + 1, handle);
      //      System.err.println("--- new: " + handle);
    }
  }

  /** limit must be within X-(X+10). Returns a handle to the next result subset of 10 elements */
  private String lookup10(int startFrom, String handle) throws SDLIPException {

    String rawSourceOutput = doSourceLookup(startFrom, handle); // startFrom
    //    System.err.println("--- " + rawSourceOutput);
    return parseResult(startFrom, rawSourceOutput);
  }


  /*---------------------------
  *doSourceLookup
  *--------------*/

  /** Does the dirty work of invoking the HTTP command
   *  for www.witchboard.com
   *  Delivers at most 10 results */

  String encode(String s) {

    return s != null ? URLEncoder.encode(s) : "";
  }

  private String doSourceLookup(int startFrom, String handle) {

    if(startFrom < 0)
      startFrom = 0;
    if(handle == null)
      handle = "";
    
    // Prepare HTTP command
    String com = "http://www.switchboard.com/bin/cgiqa.dll?SR=&MEM=1" +
      "&SK=" + handle +
      "&F=" + encode(firstName) +
      "&L=" + encode(lastName) +
      "&T=" + encode(cityName) +
      "&S=" + encode(stateName) +
      "&NN=" + (startFrom-1) +
      "&Submit=Search";

    //    System.out.println( com );

    StringBuffer result = new StringBuffer(4096);

    try {
      // Establish HTTP connecton and get results
      URL url = new URL(com);
      BufferedReader in = new BufferedReader( new InputStreamReader( url.openStream() ));

      char[] buf = new char[1024];
      int len = 0;
    
      while(/*in.ready() &&*/ (len = in.read(buf, 0, 1024)) > 0) {
	//	System.err.println("--- READ: " + len + ", " + in.ready());
	result.append(buf, 0, len);
      }
      //      System.err.println("--- READ: done");

    } catch (Exception e) {
      System.err.println(e);
    }

    return result.toString();
  }

  /*--------------------------------
  | parseResult
  -------------*/

  /**Given the string with a result returned by the Lookup command,
   *return a vector of ResultDocument that contain the info. The
   *result could be (i) no matching person found, (ii) one person's
   *full record found, or (iii) a list of people's record, if
   *the given name was not unique in the directory
   */

  /* return handle */

  String parseResult(int startFrom, String lookupResult) throws SDLIPException {

    int numOfEntries;
    int pos;
    String line;

    //    System.out.println(lookupResult);

    // Check for no entry:
    if (lookupResult.indexOf(noDocumentsStr) > -1) {
      return null; // empty
    }

    // Check whether search name was unique, and a single record was returned:
    pos = lookupResult.indexOf(oneDocumentStr);
    if (pos != -1) {
      // Name was unique, one full record is in the string:
      parseUniqueNameResult(startFrom, lookupResult.substring(pos));
      return null;
    }

    // Must be multiple records found. Find out how many:
    pos = lookupResult.indexOf(manyDocumentStr);
    if (pos != -1) {
      // Multiple records, records are in the string:
      parseMultipleNameResult(startFrom, lookupResult.substring(pos));
      // extract handle
      int p1 = lookupResult.indexOf("Next Matches");
      if(p1 < 0)
	return null;
      int p2 = lookupResult.lastIndexOf("?SK=", p1);
      int p3 = lookupResult.indexOf('&', p2);
      return lookupResult.substring(p2 + 4, p3);
    }

    // panic, couldn't match any of the strings.  Source must have
    // change format.  Return an empty vector
    errorParsing();
    //    System.out.println("Error, results were not recognized");
    // never reached
    return null;
  }


  void errorParsing() throws SDLIPException {
    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "LSP: could not screen-scrape switchboard.com");
  }

  /*--------------------------------
  | parseUniqueNameResult
  -----------------------*/

  /**Given a full record result string obtained from a unique lookup name,
   *parse the record, creating a document result object. Return a vector
   with the object.
   */

  void parseUniqueNameResult(int startFrom, String lookupResult) throws SDLIPException {

    int start;
    int end;
    int pos;

    //    try {

    Document doc = XMLObject.getDocument();
    Element eDoc = doc.createElement(SDLIP.doc);
    Element eDID = DOMUtil.createTextNode(doc, SDLIP.DID, String.valueOf(startFrom));
    eDoc.appendChild(eDID);
    Element ePropList = doc.createElement(SDLIP.propList);
    eDoc.appendChild(ePropList);

      // find name
      start = lookupResult.indexOf("<B>");
      if (start == -1) errorParsing();
      start +=4;
      end = lookupResult.indexOf("</B>",start);
      if (end == -1) errorParsing();
      String name = lookupResult.substring(start,end);
      name = replace( name, "&nbsp;", " " );
      //      resDoc.setProperty("name", name );

      ePropList.appendChild( DOMUtil.createTextNode(doc, NAME, name) );
      //      System.err.println("NAME: " + name);
      
      // find address
      start = lookupResult.indexOf("SIZE=\"2\">", end);
      if (start == -1) errorParsing();
      start +=9;
      end = lookupResult.indexOf("<BR>",start);
      if (end == -1) errorParsing();
      //      resDoc.setProperty("address", lookupResult.substring(start,end));
      String address = lookupResult.substring(start,end);
      address = replace( address, "&#160;", " " );
      address = replace( address, "<BR>", " " );
      address = address.trim();
      ePropList.appendChild( DOMUtil.createTextNode(doc, ADDRESS, address) );
      //      System.err.println("ADDRESS: " + );

      // find phone
      start = lookupResult.indexOf("<NOBR>", end);
      if (start == -1) errorParsing();
      start +=6;
      end = lookupResult.indexOf("</B>",start);
      if (end == -1) errorParsing();
      //      resDoc.setProperty("phone", lookupResult.substring(start,end));
      ePropList.appendChild( DOMUtil.createTextNode(doc, PHONE, lookupResult.substring(start,end)) );

      // add the result to the vector
      //      resVector.addElement( resDoc );

      //    } catch (ErrorParsing e) {
      //      System.out.println("Error parsing unique entry");
      //    } 
      records.addElement(eDoc);
  }

  void parseMultipleNameResult(int current, String lookupResult) throws SDLIPException {

    int start;
    int end;
    int pos;
    
    //    int current = range.get();

    Document doc = XMLObject.getDocument();
    //    try {
      // look for segment of the string were results are
      start = lookupResult.indexOf("<TR align=left");
      end = lookupResult.indexOf("COLSPAN=7");
      if ((start == -1)||(end == -1)) errorParsing();
      lookupResult = lookupResult.substring(start,end);

      // look for start of first name and iterate over all results
      start = lookupResult.indexOf("<B>");

      while (start != -1) {

	//create new document
	Element eDoc = doc.createElement(SDLIP.doc);
	Element eDID = DOMUtil.createTextNode(doc, SDLIP.DID, String.valueOf(current));
	eDoc.appendChild(eDID);
	Element ePropList = doc.createElement(SDLIP.propList);
	eDoc.appendChild(ePropList);
	  
	// find name
	start +=3;
	end = lookupResult.indexOf("</B>",start);

	//	if(

	if (end == -1) errorParsing();
	String name = lookupResult.substring(start,end);
	//	resDoc.setProperty("name", name );
	ePropList.appendChild( DOMUtil.createTextNode(doc, NAME, name) );
	// System.err.println("NAME: " + name);
	
	// find address
	start = lookupResult.indexOf("<BR>\n", end);
	if (start == -1) errorParsing();
	start +=5;
	end = lookupResult.indexOf("</font>",start);
	if (end == -1) errorParsing();
	String address = lookupResult.substring(start,end);
	address = replace( address, "&#160;", " " );
	address = replace( address, "<BR>", " " );
	address = address.trim();
	// resDoc.setProperty("address", address);
	ePropList.appendChild( DOMUtil.createTextNode(doc, ADDRESS, address) );
	//	System.err.println("ADDRESS: " + address);

	// find phone
	int m = lookupResult.indexOf("Email", end);
	start = lookupResult.indexOf("<b>", end);

	if(start > 0 && start < m) {

	  start +=3;
	  if (start == -1) errorParsing();
	  end = lookupResult.indexOf("</B>",start);
	  if (end == -1) errorParsing();
	  // resDoc.setProperty("phone", lookupResult.substring(start,end));
	  ePropList.appendChild( DOMUtil.createTextNode(doc, PHONE, lookupResult.substring(start,end)) );
	  //	System.err.println("PHONE: " + lookupResult.substring(start,end));
	} else {
	  //	  System.err.println("--- MISSING phone");
	}

	//	System.err.println("--- C: " + current + ", R: " + range.get());

	records.addElement(eDoc);
	//	System.err.println("--- ADDING: " + current + ", S=" + start + ", E=" + end);
	current++;

	// find next name (if any)
	start = lookupResult.indexOf("<B>", end);
	//	System.err.println("--- NEXT: " + start);
      }
      //    } catch (ErrorParsing e) {
      //      System.out.println("Error parsing multiple entries");
      //    } 
  }

  private String replace( String src, String in, String out ) {
    int pos = src.indexOf(in);
    
    while (pos != -1) {
      src = src.substring(0,pos) + out + src.substring(pos+in.length());
      pos = src.indexOf(in);
    }   
    return src;
  }

  public static void main(String[] args) throws Exception {

    if(args.length == 0) {
      System.err.println("Run sdlip.examples.SwitchBoardLSP! This class is used for debugging only.\n" +
			 "Usage: java.examples.SwitchBoard <first selection> [<second selection>]");
      System.exit(1);
    }

    Element eSearchResult = XMLObject.getDocument().createElement(SDLIP.SearchResult);
    SwitchBoard b = new SwitchBoard(args[0]);

    b.getDocs(eSearchResult, new RangeEnumerator(args[1]), null);
    System.out.println( DOMWriter.prettyPrint(eSearchResult) );

    if(args.length > 2) {
      eSearchResult = XMLObject.getDocument().createElement(SDLIP.SearchResult);
      b.getDocs(eSearchResult, new RangeEnumerator(args[2]), null);
      System.out.println( DOMWriter.prettyPrint(eSearchResult) );
    }
  }
}
