/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package sdlip.helpers;

import sdlip.DAV;
import sdlip.Search;
import sdlip.ResultAccess;
import sdlip.SDLIP;
import sdlip.SDLIPException;
import sdlip.http.*;
import sdlip.xml.dom.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.omg.CORBA.IntHolder;

import java.net.*;
import java.io.*;
import java.util.*;

public class ClientDaslTransport extends ClientTransportModule {

  static final int MAX_HOP_COUNT = 3; // allow for maximum 3 native DASL redirections
  static final int DEFAULT_SEARCH_TIMEOUT_REP = 0;
  static final int DEFAULT_SID_REP = 0;

  String HTTP_SEARCH_METHOD = "SEARCH";
  String HTTP_PROTOCOL_VERSION = "1.1";

  URL lsp;
  HTTPClient client;
  // HTTP basic authorization
  String authorization;

  public ClientDaslTransport() {
  }

  public ClientDaslTransport(String lsp) throws SDLIPException {

    setLspUri(lsp);
  }

  public void setLspUri(String lspStr) throws SDLIPException {

    try {
      if(lspStr.startsWith("dasl:"))
	lspStr = lspStr.substring(5);
      lsp = new URL(lspStr);
      if("http".equals(lsp.getProtocol()))
	return; // OK
    } catch (MalformedURLException mue) {}
    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Unsupported protocol mapping: " + lsp);
  }

  // Glyphica access tested at http://206.86.239.20:2518/
  //       ((ClientDaslTransport)tm).setProtocolVersion("1.0");
  //       ((ClientDaslTransport)tm).setAuthorization("webdav", "rfc2518");

  /** experimental use only; needed to talk to Glyphica server */
  public void setAuthorization(String name, String passwd) {

    this.authorization = new org.w3c.tools.codec.Base64Encoder(name + ":" + passwd).processString();
  }

  /** experimental use only; needed to talk to Glyphica server */
  public void setProtocolVersion(String version) {
    HTTP_PROTOCOL_VERSION = version;
  }

  void init() {
    client = new HTTPClient(lsp);
  }

  /**
   * Implementation of the search method of the Abstract Interface
   */
  public void search(
		     int clientSID,
		     sdlip.XMLObject subcols,
		     sdlip.XMLObject query,
		     int numDocs,
		     sdlip.XMLObject docProps,
		     int stateTimeoutReq,
		     sdlip.XMLObject queryOptions,
		     IntHolder expectedTotal,
		     IntHolder stateTimeout,
		     IntHolder serverSID,
		     sdlip.XMLObject serverDelegate,
		     sdlip.XMLObject result
		     )
    throws SDLIPException {

    try {
      /* get Elements from anonymous XMLObjects */
      Element eSubcols = XMLObject.create(subcols).getElement();
      Element eQuery = XMLObject.create(query).getElement();
      Element eDocProps = XMLObject.create(docProps).getElement();
      Element eQueryOptions = XMLObject.create(queryOptions).getElement();

      /* DOM document for creation of elements */
      Document doc = XMLObject.getDocument();

      /* DAV:searchrequest */
      Element searchrequest = doc.createElement(DAV.searchrequest);

      /* here go SDLIP specific elements like clientSID, stateTimeoutReq and queryOptions */
      Element sdlipSpecific = null;

      if(eQuery == null)
	throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "No query given");

      if(DAV.basicsearch.equals(eQuery.getTagName())) {

	/* Case 1: query type is DAV:basicsearch */
	searchrequest.appendChild(eQuery);
	/* SDLIP specific parameters are stored in the query */
	sdlipSpecific = eQuery;

	Element eFromClause = DOMUtil.getChild(eQuery, DAV.from);
	Element eSelectClause = DOMUtil.getChild(eQuery, DAV.select);
	Element eLimitClause = DOMUtil.getChild(eQuery, DAV.limit);

	if(eFromClause == null && eSelectClause == null && eLimitClause == null) {

	  // FIXME: nor DAV:limit?!
	  /* Case 1a: DAV:basicsearch, no DAV:select, DAV:from, DAV:limit clauses */

	  if(eSubcols != null) {
	    /* translate subcols to DAV:from */
	    eFromClause = doc.createElement(DAV.from);
	    NodeList list = eSubcols.getElementsByTagName("*");
	    for(int i = 0; i < list.getLength(); i++) {
	      Element col = (Element)list.item(i);
	      /* create DAV:scope for every collection */
	      Element scope = doc.createElement(DAV.scope);
	      //	      Element href = doc.createElement(DAV.href);
	      //	      scope.appendChild(href);
	      eFromClause.appendChild(scope);
	      if(SDLIP.subcolName.equals(col.getTagName())) {
		/* add subcollection URI */
		scope.appendChild( DOMUtil.createTextNode(doc, DAV.href, DOMUtil.getText(col)) );
	      } else if(SDLIP.resSet.equals(col.getTagName())) {
		/* add server SID using Interlib.SID prefix */
		scope.appendChild( DOMUtil.createTextNode(doc, DAV.href, SDLIP.SID + DOMUtil.getText(col)) );
	      } else
		throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "Invalid format of subcol parameter in search method");
	      scope.appendChild( DOMUtil.createTextNode(doc, DAV.depth, "infinity") );
	    }
	    /* insert DAV:from at the top of DAV:basicsearch */
	    eQuery.insertBefore(eFromClause, eQuery.getFirstChild());
	  } /* else no DAV:from clause if no subcols given */

	  eSelectClause = doc.createElement(DAV.select);
	  if(eDocProps != null) {
	    /* translate docPropList to DAV:select */
	    Element prop = doc.createElement(DAV.prop);
	    eSelectClause.appendChild(prop);
	    Node p = eDocProps.getFirstChild();
	    while(p != null) {
	      prop.appendChild(p.cloneNode(true));
	      p = p.getNextSibling();
	    }
	  } else {
	    /* retrieve all properties */
	    eSelectClause.appendChild(doc.createElement(DAV.allprop));
	  }
	  /* insert DAV:select at the top of DAV:basicsearch */
	  eQuery.insertBefore(eSelectClause, eQuery.getFirstChild());

	  /* DAV:limit is created using numDocs parameter */
	  if(numDocs != sdlip.Search.RETURN_ALL_DOCS) {
	    Element eLimit = doc.createElement(DAV.limit);
	    eLimit.appendChild( DOMUtil.createTextNode(doc, DAV.nresults, String.valueOf(numDocs)) );
	    eQuery.appendChild(eLimit);
	  }

	} else {
	  /* Case 1b: DAV:basicsearch, either select or from clause is present
	   * copy the query, ignore corresponding search parameters - done */
	}
      } else {
	/* Case 2: query type is NOT DAV:basicsearch */
	sdlipSpecific = doc.createElement(SDLIP.SearchRequest);
	searchrequest.appendChild(sdlipSpecific);
	if(eSubcols != null)
	  sdlipSpecific.appendChild(eSubcols);
	sdlipSpecific.appendChild( DOMUtil.createTextNode(doc, SDLIP.numDocs, String.valueOf(numDocs)) );
	if(eDocProps != null)
	  sdlipSpecific.appendChild(eDocProps);
	/* wrap the query */
	Element customQuery = doc.createElement(SDLIP.query);
	customQuery.appendChild(eQuery);
	sdlipSpecific.appendChild(customQuery);
      }

      /* append clientSID */
      Element eClientSID = DOMUtil.createTextNode(doc, SDLIP.clientSID, String.valueOf(clientSID));
      sdlipSpecific.appendChild(eClientSID);
      
      /* append stateTimeoutReq */
      Element eStateTimeoutReq = DOMUtil.createTextNode(doc, SDLIP.stateTimeoutReq, String.valueOf(stateTimeoutReq));
      sdlipSpecific.appendChild(eStateTimeoutReq);

      /* append wrapped queryOptions */
      if(eQueryOptions != null) {
	Element wrQO = doc.createElement(SDLIP.queryOptions);
	wrQO.appendChild(eQueryOptions);
	sdlipSpecific.appendChild(wrQO);
      }

      /* check whether DAV:href was among requested properties */
      boolean DAVhrefRequested = DOMUtil.getDescendant(searchrequest, DAV.href) != null;

      /* get reply from the LSP */
      Element multistatus = getReply(searchrequest, DAV.multistatus);

      boolean firstResponse = true;
      /* assume we're talking to native DASL */
      boolean sdlipServer = false;
      /* initialize DID counter in case we're talking to DASL
       * Also used to count the number of successfully retrieved docs */
      int docCounter = 0;
      /* whether the result contains documents containing bad properties */
      boolean hasDocsWithBadProps = false;
      /* default values */
      expectedTotal.value = sdlip.Search.UNKNOWABLE;
      stateTimeout.value = DEFAULT_SEARCH_TIMEOUT_REP;
      serverSID.value = DEFAULT_SID_REP;
      /* eSearchResult hold the search result */
      Element eSearchResult = doc.createElement(SDLIP.SearchResult);

      NodeList respList = multistatus.getElementsByTagName(DAV.response);
      for(int i=0; i < respList.getLength(); i++) {
	Element response = (Element)respList.item(i);
	String hrefText = DOMUtil.getChildText(response, DAV.href);
	if(firstResponse) {
	  firstResponse = false;
	  if(SDLIP.SearchResult.equals(hrefText)) {
	    /* this is an SDLIP server */
	    sdlipServer = true;
	    /* extract expectedTotal */
	    try {
	      expectedTotal.value = Integer.parseInt(DOMUtil.getChildText(response,  SDLIP.expectedTotal));
	    } catch (Exception any) {}
	    /* extract stateTimeout */
	    try {
	      stateTimeout.value = Integer.parseInt(DOMUtil.getChildText(response,  SDLIP.stateTimeout));
	    } catch (Exception any) {}
	    /* extract serverSID */
	    try {
	      serverSID.value = Integer.parseInt(DOMUtil.getChildText(response, SDLIP.serverSID));
	    } catch (Exception any) {}
	    /* extract serverDelegate */
	    XMLObject.setElement(serverDelegate, DOMUtil.getChild(response, SDLIP.redirect));
	  }
	} else {
	  /* this is a regular response, translate it into doc */

	  /* find propstat having status HTTP/1.1 200 */
	  Element propstat = null;
	  NodeList propstatList = response.getElementsByTagName(DAV.propstat);
	  for(int j=0; j < propstatList.getLength(); j++) {
	    Element tmpPropstat = (Element)propstatList.item(j);
	    String statusText = DOMUtil.getChildText(tmpPropstat, DAV.status);
	    if(statusText != null && statusText.length() >= 12 && "200".equals(statusText.substring(9, 12))) {
	      /* found status OK */
	      propstat = tmpPropstat;
	      break;
	    } else
	      hasDocsWithBadProps = true;
	  }
	  if(propstat == null) {
	    /* no propstat which is OK, ignore this response */
	    continue;
	  }

	  Element propList = DOMUtil.getChild(propstat, DAV.prop);
	  if(propList == null)
	    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "No DAV:prop inside DAV:propstat in response with DAV:href=" + hrefText);

	  Element eDoc = doc.createElement(SDLIP.doc);
	  Element eProps = doc.createElement(SDLIP.propList);
	  eDoc.appendChild(eProps);

	  /* copy property values; if none, ignore the response */
	  boolean hasProperties = false;

	  /* special case: DAV:href */
	  if(DAVhrefRequested && !sdlipServer) {
	    hasProperties = true;
	    eProps.appendChild(DOMUtil.createTextNode(doc, DAV.href, hrefText));
	  }
	  /* regular properties */
	  Node p = propList.getFirstChild();
	  while(p != null) {
	    if(p.getNodeType() == Node.ELEMENT_NODE) {
	      hasProperties = true;
	      eProps.appendChild(p.cloneNode(true));
	    }
	    p = p.getNextSibling();
	  }

	  if(!hasProperties) {
	    /* empty response, ignore it */
	    continue;
	  }

	  /* well-formed response, add to it to the result */
	  eSearchResult.appendChild(eDoc);
	  docCounter++;

	  /* do we have a DID? */
	  Element eDID = null;
	  if(sdlipServer) {
	    eDID = DOMUtil.getChild(response, SDLIP.DID);
	    if(eDID == null)
	      throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "DID element missing in one of responses of an SDLIP server reply");
	  } else
	    eDID = DOMUtil.createTextNode(doc, SDLIP.DID, String.valueOf(docCounter));
	  eDoc.insertBefore(eDID, eDoc.getFirstChild());

	} /* end of regular response */

      } /* end of response loop */

      if(docCounter == 0 && hasDocsWithBadProps)
	/* no valid responses, throw an error */
	throw new SDLIPException(SDLIPException.INVALID_REQUEST_EXC, "Native DASL result contains only bad properties");
      // DASL servers always deliver everything they have...
      if(!sdlipServer)
	expectedTotal.value = docCounter;

      XMLObject.setElement(result, eSearchResult);

    } catch (Exception any) {

      dbg.println(any);

      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }

  int getResponseCode(String line) {
    
    try {
      StringTokenizer st = new StringTokenizer(line);
      st.nextToken();
      return Integer.parseInt(st.nextToken());
    } catch (Exception any) {}
    return -1;
  }

  Element getReply(Element req, String expectedReplyTag) throws Exception {

    Element reply = getReply(req, expectedReplyTag != null);
    if(expectedReplyTag != null && !expectedReplyTag.equals(reply.getTagName()))
      throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Invalid reply: top XML element " + expectedReplyTag + " expected");
    else
      return reply;
  }

  Element getReply(Element req, boolean hasReply) throws Exception {

    int hopCount = 0;
    /* redirection loop; make sure we do not enter a dead loop
     * on native (transparent) DAV redirections */
    while(hopCount++ <= MAX_HOP_COUNT) {

      dbg.println(DBG.SHORT, toString() + " sending request to " + lsp);
      init();

      PrintWriter pw = new PrintWriter(new OutputStreamWriter(client.getOutputStream(), "UTF8"));
      //      PrintStream pw = new PrintStream(client.getOutputStream());

      String str = DOMWriter.serialize(req);

      if(dbg.getLevel() >= DBG.VERBOSE)
	dbg.println(DBG.VERBOSE, "--- SENDING: \n" + DOMWriter.prettyPrint(req));

      //      if("on".equals(System.getProperty("sdlip.debug")))
      //	pw = new PrintWriter(new OutputStreamWriter(System.out));

      pw.println(HTTP_SEARCH_METHOD + " " + lsp.getFile() + " HTTP/" + HTTP_PROTOCOL_VERSION);
      pw.println("Content-Encoding: utf-8");
      pw.println("Content-Length: " + str.length()); // HTTPUtil.utfLength(str));
      pw.println("Content-Type: text/xml");
      // FIXME: remove this line
      if(authorization != null)
	pw.println("Authorization: Basic " + authorization); // d2ViZGF2OnJmYzI1MTg=");
      pw.println();
      
      pw.print(str);
      pw.flush();

      try {
	if("on".equals(System.getProperty("sdlip.debug"))) {
	  System.out.println("--- RECEIVED:");
	  SM.util.FileUtils.copyStream(client.getInputStream(), System.out);
	}
      } catch (Exception any ) {
	// e.g. applet security exception
      }
      
      /* we are synchronous, so get the result now */
      BufferedReader din = new BufferedReader(new InputStreamReader(client.getInputStream()));
      String response = din.readLine();
      int responseCode = getResponseCode(response);

      /* valid response code? */
      switch(responseCode) {
      case DAV.rMultiStatus:
      case DAV.eBadRequest:
      case DAV.eUnprocessableEntity:
      case DAV.eInsufficientStorage: break;
      default:
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC,
				 response == null ? "Bad SDLIP server" : "Invalid DASL response: " + response);
      }

      /* read header */
      Hashtable header = HTTPUtil.getMIMEHeader(din);
      String contentType = (String)header.get("content-type");
      if(contentType != null &&
	 !("text/xml".equals(contentType) || "application/xml".equals(contentType)))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Cannot handle server response content type \"" + contentType + "\"");

      /* some SDLIP methods have to return value */
      if(!hasReply && responseCode == DAV.rMultiStatus)
	return null;

      /* read XML */
      Element el = new XMLObject( new InputSource(din) ).getElement();

      if(dbg.getLevel() >= DBG.VERBOSE)
	dbg.println(DBG.VERBOSE, "--- RECEIVED: \n" + DOMWriter.prettyPrint(el));

      // check whether an error occurred, if any
      Element eErrorResponse = DOMUtil.getDescendant(el, SDLIP.exceptionCode);
      if(eErrorResponse != null) {
	String errorCode = DOMUtil.getText(eErrorResponse);
	String desc = DOMUtil.getDescendantText(el, DAV.responsedescription);
	Element ePropList = DOMUtil.getChild((Element)eErrorResponse.getParentNode(), SDLIP.propList);
	throw new SDLIPException((short)Integer.parseInt(errorCode),
				 desc,
				 ePropList != null ? new XMLObject(ePropList) : null);
      }

      /* native DASL redirection? */
      if(responseCode == DAV.eBadRequest) {
	/* take the first redirect arbiter */
	String redirectArbiter = DOMUtil.getDescendantText(el, DAV.redirectarbiter);
	if(redirectArbiter != null) {
	  /* must be a URL, since we're talking to native DASL */
	  lsp = new URL(redirectArbiter);
	  continue;
	}
      }

      /* no SDLIP error code, but still an error: native DASL service
       * recall: InsufficientStorage is a successfull reply in SDLIP */
      if(responseCode != DAV.rMultiStatus && responseCode != DAV.eInsufficientStorage) {
	String desc = DOMUtil.getDescendantText(el, DAV.responsedescription);
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, desc);
      }

      return el;
    }
    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Maximal hop count reached");
  }

  public void cancelRequest(
			    int serverSID,
			    int reqID
			    )
    throws SDLIPException {

    try {
      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.CancelRequest);
      
      /* insert serverSID */
      Element eServerSID = DOMUtil.createTextNode(doc, SDLIP.serverSID, String.valueOf(serverSID));
      req.appendChild(eServerSID);

      /* insert reqID */
      Element eReqID = DOMUtil.createTextNode(doc, SDLIP.reqID, String.valueOf(reqID));
      req.appendChild(eReqID);

      /* no results expected */
      getReply(req, false);
      
    } catch (Exception any) {

      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }


  public void getSessionInfo(
			     int serverSID,
			     IntHolder expectedTotal,
			     IntHolder stateTimeout
			     )
    throws sdlip.SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.GetSessionInfo);
      
      /* insert serverSID */
      Element eServerSID = DOMUtil.createTextNode(doc, SDLIP.serverSID, String.valueOf(serverSID));
      req.appendChild(eServerSID);

      /* get reply */
      Element reply = getReply(req, SDLIP.SessionInfo);

      /* extract expectedTotal */
      expectedTotal.value = Integer.parseInt(DOMUtil.getChildText(reply, SDLIP.expectedTotal));
      
      /* extract stateTimeout */
      stateTimeout.value = Integer.parseInt(DOMUtil.getChildText(reply, SDLIP.stateTimeout));

    } catch (Exception any) {

      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
    
  }
  
  public void getDocs(
		      int serverSID,
		      int reqID,
		      sdlip.XMLObject docPropList,
		      String docsToGet,
		      sdlip.XMLObject res )
    throws sdlip.SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.GetDocs);

      /* insert serverSID */
      Element eServerSID = DOMUtil.createTextNode(doc, SDLIP.serverSID, String.valueOf(serverSID));
      req.appendChild(eServerSID);

      /* insert reqID */
      Element eReqID = DOMUtil.createTextNode(doc, SDLIP.reqID, String.valueOf(reqID));
      req.appendChild(eReqID);

      /* insert propList */
      Element ePropList = XMLObject.create(docPropList).getElement();
      if(ePropList != null)
	req.appendChild(ePropList);
      /* else: all properties */

      /* insert docSelection */
      req.appendChild( DOMUtil.createTextNode(doc, SDLIP.docSelection, docsToGet) );

      /* get reply */
      Element reply = getReply(req, SDLIP.SearchResult);
      
      XMLObject.setElement(res, reply);

    } catch (Exception any) {

      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }      
  }
  
  public void extendStateTimeout(
				 int serverSID,
				 int additionalTime,
				 IntHolder timeAllotted )
    throws sdlip.SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.ExtendStateTimeout);

      /* insert serverSID */
      Element eServerSID = DOMUtil.createTextNode(doc, SDLIP.serverSID, String.valueOf(serverSID));
      req.appendChild(eServerSID);

      /* insert stateTimeoutReq */
      Element eTReq = DOMUtil.createTextNode(doc, SDLIP.stateTimeoutReq, String.valueOf(additionalTime));
      req.appendChild(eTReq);

      /* get reply */
      Element reply = getReply(req, SDLIP.SessionInfo);

      /* extract stateTimeout */
      timeAllotted.value = Integer.parseInt(DOMUtil.getChildText(reply, SDLIP.stateTimeout));

    } catch (Exception any) {

      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }

  /*
  public void removeDocs(
			 int serverSID,
			 String docsToRemove )
    throws sdlip.SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.RemoveDocs);

      / * insert serverSID * /
      Element eServerSID = DOMUtil.createTextNode(doc, SDLIP.serverSID, String.valueOf(serverSID));
      req.appendChild(eServerSID);

      / * insert docSelection * /
      req.appendChild( DOMUtil.createTextNode(doc, SDLIP.docSelection, docsToRemove) );

      / * no reply expected * /
      getReply(req, false);
      
    } catch (Exception any) {

      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }      

  }
  */

  // Metadata interface

  public void getInterface(sdlip.XMLObject intf) throws SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.GetInterface);

      /* get reply */
      Element reply = getReply(req, SDLIP.SDLIPInterface);

      /* extract interfaces */
      XMLObject.setElement(intf, reply);
      
    } catch (Exception any) {
      
      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }

  public void getSubcollectionInfo(sdlip.XMLObject subcolInfo) throws SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.GetSubcollectionInfo);

      /* get reply */
      Element reply = getReply(req, SDLIP.subcolInfo);

      /* extract subcolInfo */
      XMLObject.setElement(subcolInfo, reply);

    } catch (Exception any) {
      
      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }

  public void getPropertyInfo(String subcolName, 
			      sdlip.XMLObject propInfo) throws SDLIPException {

    try {

      Document doc = XMLObject.getDocument();
      Element req = doc.createElement(SDLIP.GetPropertyInfo);

      /* insert subcolName */
      req.appendChild( DOMUtil.createTextNode(doc, SDLIP.subcolName, subcolName) );

      /* get reply */
      Element reply = getReply(req, SDLIP.propList);

      /* extract propInfo */
      XMLObject.setElement(propInfo, reply);

    } catch (Exception any) {
      
      dbg.println(any);
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }

  public String toString() {
    return "[SDLIP/DASL Client]";
  }

  static {
    ClientTransportModule.register("dasl", "sdlip.helpers.ClientDaslTransport");
    ClientTransportModule.register("http", "sdlip.helpers.ClientDaslTransport");
  }
}
