package sdlip.helpers;

import java.util.Vector;

/**
 * Timer.
 *
 * Usage:
 * <ol>
 * <li> "this" which must implement Timeoutable will
 *      be called in 3 sec, a separate thread will be
 *      created to accomplish this call
 * 
 *    Timer mytimer = new Timer(3, this, true);
 * </li>
 *
 * <li> Same as above, but no extra thread will be
 *      created for the timeout call.
 *      Make sure that this call returns quickly.
 * <pre>
 *    Timer mytimer = new Timer(3, this, false);
 * </pre>
 * </li>
 *
 * <li> Timer without a Timeoutable: no callback, if
 *      mytimer.getTimeout() == 0 it is done.
 * <pre>
 *    Timer mytimer = new Timer(3);
 * </pre>
 * </li>
 *
 * </ol>
 *
 * In the second example a single thread serves all Timeoutable's.
 * First example is a bit slower but more robust.
 */

public class Timer extends Thread {

  /** User timer. Callback called within the global timer thread */
  public static final int REGULAR = 0;

  /** User timer. Callback called within an extra thread. */
  public static final int ROBUST = 1;

  /** This timer monitors user timers;
   * single instance stored in static variable monitor */
  static final int MONITOR = 2;

  /** This is a timer which is started to
   * callback the timeout method if needed */
  static final int KILLER = 3;

  /** Sleep interval for the monitor is 1 sec */
  static final int SLEEP = 1000;

  /** Instance variables */
  long time;
  Timeoutable timeoutable;
  int type;
  Object handle;
  String name;

  /** Vector of all Timers */
  static Vector timers = new Vector();
  /** The running Timer */
  static Timer monitor = new Timer(0, null, null, MONITOR);
  /** Timer ID */
  static int timerID = 0;

  /** Returns remaining time until timeout method is called.
   *  Zero if timed out. */
  public synchronized int getTimeout() {

    long current = System.currentTimeMillis();
    if(time >= current)
      return (int)(time - current) / 1000;
    else
      return 0;
  }

  public boolean isTimedOut() {
    return getTimeout() <= 0;
  }

  /** Sets a new timeout */
  public synchronized void setTimeout(int reltime) {

    setAbsolute(System.currentTimeMillis() + reltime * 1000);
  }

  /** Sets absolute time for timeout */
  public synchronized void setAbsolute(long abstime) {

    time = abstime;
  }

  /**
   * General user timer constructor.
   * timerType is either REGULAR or ROBUST.
   */
  public Timer(int reltime, Timeoutable to, Object handle, int timerType) {

    this();
    timeoutable = to;
    type = timerType;
    this.handle = handle;
    //    System.err.println("--- Creating timer with handle: " + handle);
    setTimeout(reltime);
    if(type == ROBUST || type == REGULAR)
      register(this);
    // note that only the MONITOR and KILLER are actually started
    setPriority(Thread.NORM_PRIORITY + 1);
  }

  /**
   * Creates a "robust" timer.
   */
  public Timer(int reltime, Timeoutable to, Object handle) {
    this(reltime, to, handle, ROBUST);
  }

  /**
   * Creates a timer which has no Timeoutable.
   */
  public Timer(int reltime) {
    this(reltime, null, null, REGULAR);
  }

  private Timer() {
    super();
    name = "Timer-" + ++timerID;
}

  /**
   * Delivers the Timeoutable under observation
   */
  public Timeoutable getTimeoutable() {
    return timeoutable;
  }

  /**
   * Delivers the handle that is passed to the Timeoutable
   */
  public Object getHandle() {
    return handle;
  }

  /**
   * Returns REGULAR or ROBUST for a user timer.
   */
  public int getType() {
    return type;
  }

  /** stop the timer */
  public void interrupt() {
    timers.removeElement(this);
    time = System.currentTimeMillis();
  }

  public String toString() {

    return name + "(" + type + ")";
  }

  /**
   * The running instance of the Timer. It is used for internal
   * purposes and must not be publicly accessed.
   */
  public synchronized void run() {

    if(type == KILLER) {
      // just call timeout
      timeItOut(timeoutable);
      return;
    }
    // else type == MONITOR

    while(true) {
      
      try {
	sleep(SLEEP);
      } catch (InterruptedException intex) {}
      
      // check timers
      for(int i=0; i < timers.size(); i++) {
	Timer t = (Timer)timers.elementAt(i);
	//	System.err.println("--- Checking " + t + " of " + timers.size());
	if(t.getTimeout() == 0) {
	  // call timeout
	  if(t.getType() == REGULAR)
	    // call within this thread
	    timeItOut(t.getTimeoutable());
	  else { // ROBUST
	    //	    System.err.println("--- NEW");
	    Timer temp = new Timer(0, t.getTimeoutable(), t.getHandle(), KILLER);
	    temp.start();
	  }
	  // remove timer from the list
	  timers.removeElementAt(i--);
	}
      }
      // check whether no timers running
      synchronized(timers) {
	if(timers.size() == 0) {
	  // prepare another monitor
	  monitor = new Timer(0, null, null, MONITOR);
	  // stop the monitor
	  break;
	}
      }
    }
  }

  void timeItOut(Timeoutable t) {
    try {
      t.timeout(handle);
    } catch (Exception any) {
      System.err.println("Timeouting of " + t + " caused " + any + " in the thread " + this);
    }
  }

  static void register(Timer userTimer) {

    // System.err.println("--- Registering " + userTimer);

    // just make sure that this particular method is synchronized,
    // not the whole monitor instance
    synchronized(timers) {
      
      if(!monitor.isAlive()) {
	//	System.err.println("--- NEW MONITOR");
	monitor.start();
      }
      timers.addElement(userTimer);
    }
  }

  public static void main(String[] args) {

    for(int i=0; i < args.length; i++)
      new Timer(Integer.parseInt(args[i]), null, null, Timer.ROBUST);

    try {
      Thread.sleep(10000);
    } catch (InterruptedException ie) {}

    for(int i=0; i < args.length; i++)
      new Timer(Integer.parseInt(args[i]), null, null, Timer.ROBUST);
  }

}
