package sdlip.xml.dom;
                    
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import org.w3c.dom.*;

import org.xml.sax.*;
import java.net.URL;
import java.util.*;
import java.io.*;

/**
 */
public class DOMWriter {

  /** Print writer. */
  protected PrintWriter out;

  /** Canonical output. */
  protected boolean canonical;

  boolean autoIndent = false;

  private boolean insertNamespaces = true;
  private Hashtable namespaces = null;

  // Constructors

  /** Default constructor. */
  public DOMWriter(PrintWriter out)  {
    this(out, false);
  }

  public DOMWriter(boolean canonical)  {
    
    this(new PrintWriter(new OutputStreamWriter(System.out)), canonical);
  }

  public DOMWriter()  {
    
    this(false);
  }
  
  protected DOMWriter(PrintWriter out, boolean canonical)  {
    
    this.out = out;
    this.canonical = canonical;
    
  }

  // Public static methods
  
  /** Prints the resulting document tree. */
  public static void print(String uri, boolean canonical) {
    
    try {
      DOMWriter writer = new DOMWriter(canonical);
      InputSource source = new InputSource( new URL(uri).openStream() );
      writer.print(new DOMParser().parse(source));
    } catch (SAXException se) {
      se.getException().printStackTrace(System.err);
    } catch (Exception e) {
      System.err.println("DOMWriter: " + e.getClass());
      e.printStackTrace(System.err);
    }
    
  } // print(String,String,boolean)
  
  public static String prettyPrint(String xmldata) throws Exception {
    
    InputSource source = new InputSource( xmldata );
    return DOMWriter.prettyPrint(new DOMParser().parse(source));
  }
  
  public static String prettyPrint(Node node) {

    if(node == null)
      return null;

    StringWriter sw = new StringWriter();
    DOMWriter writer = new DOMWriter(new PrintWriter(sw, true));
    writer.setAutoIndent(true);
    writer.print(node);
    return sw.toString();
  }
  
  public void setAutoIndent(boolean b) {
    autoIndent = b;
  }

  public static String serialize(Node node) {

    if(node == null)
      return null;

    StringWriter sw = new StringWriter();
    DOMWriter writer = new DOMWriter(new PrintWriter(sw, true));
    writer.print(node);
    return sw.toString();
  }

    //
    // Public methods
    //
    public void print(Node node) {
      noIndent = true;
      print(node, 0);
    }

  boolean noIndent = false;
  boolean firstIndent = true;
  
  void indent(int depth) {
    
    if(autoIndent && !firstIndent) {
      out.print('\n');
      while(depth-- > 0)
	out.print('\t');
    }
    firstIndent = false;
  }
  
  /** Prints the specified node, recursively. */
  void print(Node node, int depth) {
    
    // is there anything to do?
    if (node == null) {
      return;
    }
    
    int type = node.getNodeType();
    switch (type) {
      // print document
    case Node.DOCUMENT_NODE: {
      if (!canonical) {
	//                    out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
	out.println("<?xml version=\"1.0\"?>");
      }
      print(((Document)node).getDocumentElement());
      out.flush();
      break;
    }
    
    // print element with attributes
    case Node.ELEMENT_NODE: {
      
      if(namespaces == null) {
	namespaces = ((DocumentImpl)node.getOwnerDocument()).getNamespaces((Element)node);
	//		  out.print("<?xml version=\"1.0\"?>");
      }
      
      indent(depth);
      out.print('<');
      String name = ((DocumentImpl)node.getOwnerDocument()).getShort(namespaces, node.getNodeName());
      out.print(name);
      
      if(insertNamespaces) {
	insertNamespaces = false;
	for(Enumeration en = namespaces.keys(); en.hasMoreElements();) {
	  String ns = (String)en.nextElement();
	  String abb = (String)namespaces.get(ns);
	  out.print(" xmlns:");
	  out.print(abb);
	  out.print("=\"");
	  out.print(ns);
	  out.print('"');
	}
      }
      
      Attr attrs[] = sortAttributes(node.getAttributes());
      for (int i = 0; i < attrs.length; i++) {
	Attr attr = attrs[i];
	out.print(' ');
	// BUG: fixed on 10/07/00, reported by Noah Green, Columbia
	out.print(((DocumentImpl)node.getOwnerDocument()).getShort(namespaces, attr.getNodeName()));
	//	out.print(attr.getNodeName());
	out.print("=\"");
	out.print(normalize(attr.getNodeValue()));
	out.print('"');
      }
      NodeList children = node.getChildNodes();
      int len = children.getLength();
      if (len > 0) {
	out.print('>');
	for (int i = 0; i < len; i++) {
	  print(children.item(i), depth+1);
	}
	if(!noIndent)
	  indent(depth);
	out.print("</");
	out.print(name);
	out.print('>');
      } else
	out.print("/>");
      noIndent = false;
      break;
    }
    
    // handle entity reference nodes
    case Node.ENTITY_REFERENCE_NODE: {
      if (canonical) {
	NodeList children = node.getChildNodes();
	if (children != null) {
	  int len = children.getLength();
	  for (int i = 0; i < len; i++) {
	    print(children.item(i), depth+1);
	  }
	}
      }
      else {
	out.print('&');
	out.print(node.getNodeName());
	out.print(';');
      }
      break;
    }
    
    // print cdata sections
    case Node.CDATA_SECTION_NODE: {
      if (canonical) {
	out.print(normalize(node.getNodeValue()));
      }
      else {
	out.print("<![CDATA[");
	out.print(node.getNodeValue());
	out.print("]]>");
      }
      break;
    }
    
    // print text
    case Node.TEXT_NODE: {
      
      String str = normalize(node.getNodeValue());
      if(autoIndent) {
	str = str.trim();
	if(str.length() > 0)
	  noIndent = true;
      }
      out.print(str);
      break;
    }
    
    // print processing instruction
    case Node.PROCESSING_INSTRUCTION_NODE: {
      out.print("<?");
      out.print(node.getNodeName());
      String data = node.getNodeValue();
      if (data != null && data.length() > 0) {
	out.print(' ');
	out.print(data);
      }
      out.print("?>");
      break;
    }
    }
    
    out.flush();
    
  } // print(Node)

  // Protected methods
  
  boolean isEntityRef(String s, int pos) {
    
    // non-whitespace followed by ';'
    while(pos < s.length()) {
      char c = s.charAt(pos++);
      if(c == ';')
	return true;
      else if (Character.isWhitespace(c))
	break;
    }
    return false;
  }
  
  /** Normalizes the given string. */
  protected String normalize(String s) {
    StringBuffer str = new StringBuffer();
    
    int len = (s != null) ? s.length() : 0;
    for (int i = 0; i < len; i++) {
      char ch = s.charAt(i);
      switch (ch) {
      case '<': {
	str.append("&lt;");
	break;
      }
      case '>': {
	str.append("&gt;");
	break;
      }
      case '&': {
	if(!isEntityRef(s, i+1))
                    str.append("&amp;");
	else
		    str.append(ch);
	break;
                }
      case '"': {
	str.append("&quot;");
	break;
      }
      case '\r':
      case '\n': {
	if (canonical) {
	  str.append("&#");
	  str.append(Integer.toString(ch));
	  str.append(';');
	  break;
	}
	// else, default append char
      }
      default: {
	str.append(ch);
      }
      }
    }
    
    return str.toString();
    
  } // normalize(String):String
  
  /** Returns a sorted list of attributes. */
  protected Attr[] sortAttributes(NamedNodeMap attrs) {
    
    int len = (attrs != null) ? attrs.getLength() : 0;
    Attr array[] = new Attr[len];
    for (int i = 0; i < len; i++) {
      array[i] = (Attr)attrs.item(i);
    }
    for (int i = 0; i < len - 1; i++) {
      String name  = array[i].getNodeName();
      int    index = i;
      for (int j = i + 1; j < len; j++) {
	String curName = array[j].getNodeName();
	if (curName.compareTo(name) < 0) {
	  name  = curName;
	  index = j;
	}
      }
      if (index != i) {
	Attr temp    = array[i];
	array[i]     = array[index];
	array[index] = temp;
      }
    }
    
    return array;
    
  } // sortAttributes(NamedNodeMap):Attr[]
  
  //
  // Main
  //
  
  /** Main program entry point. */
  public static void main(String argv[]) {
    
    // is there anything to do?
    if (argv.length == 0) {
      printUsage();
      System.exit(1);
    }
    
    // vars
    boolean canonical  = false;
    
    // check parameters
    for (int i = 0; i < argv.length; i++) {
      String arg = argv[i];
      
      // options
      if (arg.startsWith("-")) {
	
	if (arg.equals("-c")) {
	  canonical = true;
	  continue;
	}
	
	if (arg.equals("-h")) {
	  printUsage();
	  System.exit(1);
	}
      }
      
      // print uri
      System.err.println(arg+':');
      print(arg, canonical);
      System.err.println();
    }
    
  } // main(String[])
  
  /** Prints the usage. */
  private static void printUsage() {
    
    System.err.println("usage: java SDLIP.util.DOMWriter (options) uri ...");
    System.err.println();
    System.err.println("options:");
    System.err.println("  -c       Canonical XML output.");
    System.err.println("  -h       This help screen.");
    
  } // printUsage()
  
} // class DOMWriter
